package org.eclipse.jdt.internal.compiler.ast;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.Statement;
import org.eclipse.jdt.internal.compiler.codegen.BranchLabel;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.LoopingFlowContext;
import org.eclipse.jdt.internal.compiler.flow.UnconditionalFlowInfo;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;

public class WhileStatement extends Statement {
  public Expression condition;
  
  public Statement action;
  
  private BranchLabel breakLabel;
  
  private BranchLabel continueLabel;
  
  int preCondInitStateIndex = -1;
  
  int condIfTrueInitStateIndex = -1;
  
  int mergedInitStateIndex = -1;
  
  public WhileStatement(Expression condition, Statement action, int s, int e) {
    this.condition = condition;
    this.action = action;
    if (action instanceof org.eclipse.jdt.internal.compiler.ast.EmptyStatement)
      action.bits |= 0x1; 
    this.sourceStart = s;
    this.sourceEnd = e;
  }
  
  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo) {
    FlowInfo actionInfo;
    UnconditionalFlowInfo unconditionalFlowInfo2;
    this.breakLabel = new BranchLabel();
    this.continueLabel = new BranchLabel();
    int initialComplaintLevel = ((flowInfo.reachMode() & 0x3) != 0) ? 1 : 0;
    Constant cst = this.condition.constant;
    boolean isConditionTrue = (cst != Constant.NotAConstant && cst.booleanValue());
    boolean isConditionFalse = (cst != Constant.NotAConstant && !cst.booleanValue());
    cst = this.condition.optimizedBooleanConstant();
    boolean isConditionOptimizedTrue = (cst != Constant.NotAConstant && cst.booleanValue());
    boolean isConditionOptimizedFalse = (cst != Constant.NotAConstant && !cst.booleanValue());
    this.preCondInitStateIndex = currentScope.methodScope().recordInitializationStates(flowInfo);
    UnconditionalFlowInfo unconditionalFlowInfo1 = flowInfo.nullInfoLessUnconditionalCopy();
    LoopingFlowContext condLoopContext;
    FlowInfo flowInfo1 = this.condition.analyseCode(
        currentScope, 
        (FlowContext)(condLoopContext = 
        new LoopingFlowContext(flowContext, flowInfo, (ASTNode)this, null, 
          null, (Scope)currentScope, true)), 
        (FlowInfo)unconditionalFlowInfo1);
    this.condition.checkNPEbyUnboxing(currentScope, flowContext, flowInfo);
    if (this.action == null || (
      this.action.isEmptyBlock() && (currentScope.compilerOptions()).complianceLevel <= 3080192L)) {
      condLoopContext.complainOnDeferredFinalChecks(currentScope, 
          flowInfo1);
      condLoopContext.complainOnDeferredNullChecks(currentScope, 
          (FlowInfo)flowInfo1.unconditionalInits());
      if (isConditionTrue)
        return (FlowInfo)FlowInfo.DEAD_END; 
      FlowInfo mergedInfo = flowInfo.copy().addInitializationsFrom(flowInfo1.initsWhenFalse());
      if (isConditionOptimizedTrue)
        mergedInfo.setReachMode(1); 
      this.mergedInitStateIndex = 
        currentScope.methodScope().recordInitializationStates(mergedInfo);
      return mergedInfo;
    } 
    LoopingFlowContext loopingContext = 
      new LoopingFlowContext(
        flowContext, 
        flowInfo, 
        (ASTNode)this, 
        this.breakLabel, 
        this.continueLabel, 
        (Scope)currentScope, 
        true);
    if (isConditionFalse) {
      unconditionalFlowInfo2 = FlowInfo.DEAD_END;
    } else {
      actionInfo = flowInfo1.initsWhenTrue().copy();
      if (isConditionOptimizedFalse)
        actionInfo.setReachMode(1); 
    } 
    this.condIfTrueInitStateIndex = 
      currentScope.methodScope().recordInitializationStates(
        flowInfo1.initsWhenTrue());
    if (this.action.complainIfUnreachable(actionInfo, currentScope, initialComplaintLevel, true) < 2)
      actionInfo = this.action.analyseCode(currentScope, (FlowContext)loopingContext, actionInfo); 
    FlowInfo exitBranch = flowInfo.copy();
    int combinedTagBits = actionInfo.tagBits & loopingContext.initsOnContinue.tagBits;
    if ((combinedTagBits & 0x3) != 0) {
      if ((combinedTagBits & 0x1) != 0)
        this.continueLabel = null; 
      exitBranch.addInitializationsFrom(flowInfo1.initsWhenFalse());
      unconditionalFlowInfo2 = actionInfo.mergedWith(loopingContext.initsOnContinue.unconditionalInits());
      condLoopContext.complainOnDeferredNullChecks(currentScope, 
          (FlowInfo)unconditionalFlowInfo2, false);
      loopingContext.complainOnDeferredNullChecks(currentScope, 
          (FlowInfo)unconditionalFlowInfo2, false);
    } else {
      condLoopContext.complainOnDeferredFinalChecks(currentScope, 
          flowInfo1);
      unconditionalFlowInfo2 = unconditionalFlowInfo2.mergedWith(loopingContext.initsOnContinue.unconditionalInits());
      condLoopContext.complainOnDeferredNullChecks(currentScope, 
          (FlowInfo)unconditionalFlowInfo2);
      loopingContext.complainOnDeferredFinalChecks(currentScope, 
          (FlowInfo)unconditionalFlowInfo2);
      loopingContext.complainOnDeferredNullChecks(currentScope, 
          (FlowInfo)unconditionalFlowInfo2);
      exitBranch
        .addPotentialInitializationsFrom(
          (FlowInfo)unconditionalFlowInfo2.unconditionalInits())
        .addInitializationsFrom(flowInfo1.initsWhenFalse());
    } 
    if (loopingContext.hasEscapingExceptions()) {
      UnconditionalFlowInfo unconditionalFlowInfo;
      FlowInfo loopbackFlowInfo = flowInfo.copy();
      if (this.continueLabel != null)
        unconditionalFlowInfo = loopbackFlowInfo.mergedWith(loopbackFlowInfo.unconditionalCopy().addNullInfoFrom((FlowInfo)unconditionalFlowInfo2).unconditionalInits()); 
      loopingContext.simulateThrowAfterLoopBack((FlowInfo)unconditionalFlowInfo);
    } 
    UnconditionalFlowInfo unconditionalFlowInfo3 = FlowInfo.mergedOptimizedBranches(
        ((loopingContext.initsOnBreak.tagBits & 
        0x3) != 0) ? 
        (FlowInfo)loopingContext.initsOnBreak : 
        flowInfo.addInitializationsFrom((FlowInfo)loopingContext.initsOnBreak), 
        isConditionOptimizedTrue, 
        exitBranch, 
        isConditionOptimizedFalse, 
        !isConditionTrue);
    this.mergedInitStateIndex = currentScope.methodScope().recordInitializationStates((FlowInfo)unconditionalFlowInfo3);
    return (FlowInfo)unconditionalFlowInfo3;
  }
  
  public void generateCode(BlockScope currentScope, CodeStream codeStream) {
    if ((this.bits & Integer.MIN_VALUE) == 0)
      return; 
    int pc = codeStream.position;
    Constant cst = this.condition.optimizedBooleanConstant();
    boolean isConditionOptimizedFalse = (cst != Constant.NotAConstant && !cst.booleanValue());
    if (isConditionOptimizedFalse) {
      this.condition.generateCode(currentScope, codeStream, false);
      if (this.mergedInitStateIndex != -1) {
        codeStream.removeNotDefinitelyAssignedVariables((Scope)currentScope, this.mergedInitStateIndex);
        codeStream.addDefinitelyAssignedVariables((Scope)currentScope, this.mergedInitStateIndex);
      } 
      codeStream.recordPositionsFrom(pc, this.sourceStart);
      return;
    } 
    this.breakLabel.initialize(codeStream);
    if (this.continueLabel == null) {
      if (this.condition.constant == Constant.NotAConstant)
        this.condition.generateOptimizedBoolean(
            currentScope, 
            codeStream, 
            null, 
            this.breakLabel, 
            true); 
    } else {
      this.continueLabel.initialize(codeStream);
      if ((this.condition.constant == Constant.NotAConstant || 
        !this.condition.constant.booleanValue()) && 
        this.action != null && 
        !this.action.isEmptyBlock()) {
        int jumpPC = codeStream.position;
        codeStream.goto_(this.continueLabel);
        codeStream.recordPositionsFrom(jumpPC, this.condition.sourceStart);
      } 
    } 
    BranchLabel actionLabel = new BranchLabel(codeStream);
    if (this.action != null) {
      actionLabel.tagBits |= 0x2;
      if (this.condIfTrueInitStateIndex != -1)
        codeStream.addDefinitelyAssignedVariables(
            (Scope)currentScope, 
            this.condIfTrueInitStateIndex); 
      actionLabel.place();
      this.action.generateCode(currentScope, codeStream);
      if (this.preCondInitStateIndex != -1)
        codeStream.removeNotDefinitelyAssignedVariables((Scope)currentScope, this.preCondInitStateIndex); 
    } else {
      actionLabel.place();
    } 
    if (this.continueLabel != null) {
      this.continueLabel.place();
      this.condition.generateOptimizedBoolean(
          currentScope, 
          codeStream, 
          actionLabel, 
          null, 
          true);
    } 
    if (this.mergedInitStateIndex != -1) {
      codeStream.removeNotDefinitelyAssignedVariables((Scope)currentScope, this.mergedInitStateIndex);
      codeStream.addDefinitelyAssignedVariables((Scope)currentScope, this.mergedInitStateIndex);
    } 
    this.breakLabel.place();
    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }
  
  public void resolve(BlockScope scope) {
    TypeBinding type = this.condition.resolveTypeExpecting(scope, (TypeBinding)TypeBinding.BOOLEAN);
    this.condition.computeConversion((Scope)scope, type, type);
    if (this.action != null)
      this.action.resolve(scope); 
  }
  
  public StringBuffer printStatement(int tab, StringBuffer output) {
    printIndent(tab, output).append("while (");
    this.condition.printExpression(0, output).append(')');
    if (this.action == null) {
      output.append(';');
    } else {
      this.action.printStatement(tab + 1, output);
    } 
    return output;
  }
  
  public void traverse(ASTVisitor visitor, BlockScope blockScope) {
    if (visitor.visit(this, blockScope)) {
      this.condition.traverse(visitor, blockScope);
      if (this.action != null)
        this.action.traverse(visitor, blockScope); 
    } 
    visitor.endVisit(this, blockScope);
  }
  
  public boolean doesNotCompleteNormally() {
    Constant cst = this.condition.constant;
    boolean isConditionTrue = !(cst != null && (cst == Constant.NotAConstant || !cst.booleanValue()));
    cst = this.condition.optimizedBooleanConstant();
    boolean isConditionOptimizedTrue = (cst == null) ? true : ((cst != Constant.NotAConstant && cst.booleanValue()));
    return ((isConditionTrue || isConditionOptimizedTrue) && (this.action == null || !this.action.breaksOut(null)));
  }
  
  public boolean completesByContinue() {
    return this.action.continuesAtOuterLabel();
  }
}
